""" Main submodule for pyintesishome2 """
import asyncio
import json
import logging
import socket

import aiohttp

from .const import API_URL, API_VER, DEVICE_INTESISHOME, INTESIS_CMD_STATUS
from .exceptions import IHAuthenticationError, IHConnectionError
from .intesisbase import IntesisBase

_LOGGER = logging.getLogger("pyintesishome2")


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-public-methods
class IntesisHome(IntesisBase):
    """pyintesishome2 cloud class"""

    def __init__(
        self,
        username,
        password,
        loop=None,
        websession=None,
        device_type=DEVICE_INTESISHOME,
    ):
        super().__init__(
            device_type=device_type,
            username=username,
            loop=loop,
            password=password,
            websession=websession,
        )
        self._api_url = API_URL[device_type]
        self._api_ver = API_VER[device_type]
        self._cmd_server = None
        self._cmd_server_port = None
        self._auth_token = None
        self._controller_id = username

    async def _parse_response(self, decoded_data):
        _LOGGER.debug("%s API Received: %s", self._device_type, decoded_data)
        resp = json.loads(decoded_data)
        # Ensure any pending command awaiters are released before invoking callbacks
        if not self._received_response.is_set():
            _LOGGER.debug("Setting _received_response event")
            self._received_response.set()

        # Parse response
        if resp["command"] == "connect_rsp":
            # New connection success
            if resp["data"]["status"] == "ok":
                _LOGGER.info("%s successfully authenticated", self._device_type)
                self._connected = True
                self._connecting = False
                self._connection_retries = 0
                await self._send_update_callback()
        elif resp["command"] == "status":
            # Value has changed
            self._update_device_state(
                resp["data"]["deviceId"],
                resp["data"]["uid"],
                resp["data"]["value"],
            )
            if resp["data"]["uid"] != 60002:
                await self._send_update_callback(
                    device_id=str(resp["data"]["deviceId"])
                )
        elif resp["command"] == "rssi":
            # Wireless strength has changed
            self._update_rssi(resp["data"]["deviceId"], resp["data"]["value"])
        else:
            _LOGGER.debug("Unexpected command received: %s", resp["command"])
        return

    async def _send_keepalive(self):
        try:
            while True:
                await asyncio.sleep(120)
                if not self._connected:
                    _LOGGER.debug("Stopping keepalive task because connection is inactive")
                    break
                if not self._writer or self._writer.is_closing():
                    _LOGGER.warning("Keepalive aborted because writer is not available")
                    break
                if not self._devices:
                    _LOGGER.debug("No devices registered; skipping keepalive ping")
                    continue
                _LOGGER.debug("sending keepalive to %s", self._device_type)
                try:
                    device_id = str(next(iter(self._devices)))
                except StopIteration:
                    _LOGGER.debug("No device id available for keepalive")
                    continue
                message = (
                    f'{{"command":"get","data":{{"deviceId":{device_id},"uid":10}}}}'
                )
                await self._send_command(message, wait_for_response=False)
        except asyncio.CancelledError:
            _LOGGER.debug("Cancelled the keepalive task")

    async def connect(self):
        """Public method for connecting to IntesisHome/Airconwithme API"""
        if self._connected or self._connecting:
            return

        self._connecting = True
        self._connection_retries = 0
        self._devices = {}
        await self._cancel_task_if_exists(self._receive_task)

        try:
            self._auth_token = await self._request_auth_token()
            await self._open_command_channel()
        finally:
            self._connecting = False

    async def _request_auth_token(self):
        try:
            return await self.poll_status()
        except (IHAuthenticationError, IHConnectionError) as exc:
            _LOGGER.error("Error connecting to IntesisHome API: %s", exc)
            raise

    async def _open_command_channel(self):
        _LOGGER.debug(
            "Opening connection to %s API at %s:%i",
            self._device_type,
            self._cmd_server,
            self._cmd_server_port,
        )

        try:
            self._reader, self._writer = await asyncio.open_connection(
                self._cmd_server, self._cmd_server_port
            )
        except (ConnectionRefusedError, Exception) as exc:  # pylint: disable=broad-except
            _LOGGER.error(
                "Connection to %s:%s failed with exception %s",
                self._cmd_server,
                self._cmd_server_port,
                exc,
            )
            self._connected = False
            return

        # pylint: disable=C0209
        auth_msg = '{"command":"connect_req","data":{"token":%s}}' % (
            self._auth_token
        )
        self._receive_task = self._event_loop.create_task(self._data_received())
        await self._send_command(auth_msg)
        self._auth_token = None
        self._keepalive_task = self._event_loop.create_task(self._send_keepalive())

    async def poll_status(self, sendcallback=False):
        """Public method to query IntesisHome for state of device.
        Notifies subscribers if sendCallback True."""
        get_status = {
            "username": self._username,
            "password": self._password,
            "cmd": INTESIS_CMD_STATUS,
            "version": self._api_ver,
        }

        status_response = None
        try:
            async with self._web_session.post(
                url=self._api_url, data=get_status
            ) as resp:
                status_response = await resp.json(content_type=None)
                _LOGGER.debug(status_response)
        except aiohttp.client_exceptions.ClientConnectorError as exc:
            raise IHConnectionError from exc
        except (aiohttp.client_exceptions.ClientError, socket.gaierror) as exc:
            self._error_message = f"Error connecting to {self._device_type} API: {exc}"
            _LOGGER.error("%s Exception. %s / %s", type(exc), repr(exc.args), exc)
            raise IHConnectionError from exc

        if status_response:
            if "errorCode" in status_response:
                self._error_message = status_response["errorMessage"]
                _LOGGER.error("Error from API %s", repr(self._error_message))
                raise IHAuthenticationError()

            config = status_response.get("config")
            if config:
                self._cmd_server = config.get("serverIP")
                self._cmd_server_port = config.get("serverPort")
                self._auth_token = config.get("token")

                _LOGGER.debug(
                    "Server: %s:%i, Token: %s",
                    self._cmd_server,
                    self._cmd_server_port,
                    self._auth_token,
                )

            # Setup devices
            for installation in config.get("inst"):
                for device in installation.get("devices"):
                    self._devices[device["id"]] = {
                        "name": device["name"],
                        "widgets": device["widgets"],
                        "model": device["modelId"],
                    }
                    _LOGGER.debug(repr(self._devices))

            # Update device status
            for status in status_response["status"]["status"]:
                device_id = str(status["deviceId"])

                # Handle devices which don't appear in installation
                if device_id not in self._devices:
                    self._devices[device_id] = {
                        "name": "Device " + device_id,
                        "widgets": [42],
                    }

                self._update_device_state(device_id, status["uid"], status["value"])

            if sendcallback:
                await self._send_update_callback(device_id=str(device_id))

        return self._auth_token

    # pylint: disable=C0209
    async def _set_value(self, device_id, uid, value):
        """Internal method to send a command to the API (and connect if necessary)"""
        message = (
            '{"command":"set","data":{"deviceId":%s,"uid":%i,"value":%i,"seqNo":0}}'
            % (device_id, uid, value)
        )
        await self._send_command(message)

    def _get_fan_map(self, device_id):
        return self.get_device_property(device_id, "config_fan_map")



""" Helpers for pyintesishome2 """


def twos_complement_16bit(val):
    """Compute Two's Complement, to represent negative temperatures"""
    if (val & (1 << 15)) != 0:
        val = val - (1 << 16)
    return val


def uint32(value):
    """uint32"""
    result = int(value) & 0xFFFF
    return result



"""Tests for pyintesishome2."""
import asyncio

import aiohttp
import pytest

from pyintesishome2 import IntesisHome, IntesisHomeLocal
from pyintesishome2.const import API_URL, DEVICE_INTESISHOME

from . import mock_aioresponse  # noqa: F401
from . import (
    MOCK_DEVICE_ID,
    MOCK_HOST,
    MOCK_PASS,
    MOCK_USER,
    MOCK_VAL_RUN_HOURS,
    cloud_api_callback,
    intesisbox_api_callback,
    local_api_callback,
)

controllers = {}


async def async_setup_controllers():
    # The aiohttp.ClientSession should be created within an async function
    session = aiohttp.ClientSession()
    loop = asyncio.get_event_loop()

    controllers["local"] = IntesisHomeLocal(
        MOCK_HOST,
        MOCK_USER,
        MOCK_PASS,
        loop=loop,
        websession=session,
    )

    # controllers["intesisbox"] = IntesisBox(
    #     MOCK_HOST,
    #     loop=loop,
    #     websession=session,
    # )

    controllers["cloud"] = IntesisHome(
        MOCK_USER,
        MOCK_PASS,
        loop=loop,
        websession=session,
        device_type=DEVICE_INTESISHOME,
    )


asyncio.run(async_setup_controllers())


@pytest.mark.parametrize("controller", controllers.values(), ids=controllers.keys())
class TestPyIntesisHome2:
    @pytest.fixture(autouse=True)
    async def _setup(self, mock_aioresponse, loop):  # noqa: F811
        mock_aioresponse.post(
            f"http://{MOCK_HOST}/api.cgi",
            callback=local_api_callback,
            repeat=True,
        )

        mock_aioresponse.post(
            f"{API_URL[DEVICE_INTESISHOME]}",
            callback=cloud_api_callback,
            repeat=True,
        )

        mock_aioresponse.post(
            MOCK_HOST,
            callback=intesisbox_api_callback,
            repeat=True,
        )

    async def test_connect(self, controller):
        result = await controller.connect()
        assert result is None

    def test_get_power_state(self, controller):
        result = controller.get_power_state(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "off"

    async def test_set_power(self, controller):
        await controller.set_power_on(MOCK_DEVICE_ID)
        await controller.set_power_off(MOCK_DEVICE_ID)

    def test_get_mode(self, controller):
        result = controller.get_mode(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "cool"

    def test_get_mode_list(self, controller):
        result = controller.get_mode_list(MOCK_DEVICE_ID)
        assert isinstance(result, list)
        assert len(result)

    async def test_set_mode(self, controller):
        await controller.set_mode_heat(MOCK_DEVICE_ID)
        await controller.set_mode_cool(MOCK_DEVICE_ID)
        await controller.set_mode_fan(MOCK_DEVICE_ID)
        await controller.set_mode_auto(MOCK_DEVICE_ID)
        await controller.set_mode_dry(MOCK_DEVICE_ID)

    def test_get_fan_speed(self, controller):
        result = controller.get_fan_speed(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "quiet"

    def test_get_fan_speed_list(self, controller):
        result = controller.get_fan_speed_list(MOCK_DEVICE_ID)
        assert isinstance(result, list)
        assert len(result)

    async def test_set_fan_speed(self, controller):
        await controller.set_fan_speed(MOCK_DEVICE_ID, "high")

    def test_has_vertical_swing(self, controller):
        result = controller.has_vertical_swing(MOCK_DEVICE_ID)
        assert isinstance(result, bool)
        assert result is True

    def test_get_vertical_swing(self, controller):
        result = controller.get_vertical_swing(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "manual2"

    async def test_set_vertical_vane(self, controller):
        await controller.set_vertical_vane(MOCK_DEVICE_ID, "manual4")

    def test_has_horizontal_swing(self, controller):
        result = controller.has_horizontal_swing(MOCK_DEVICE_ID)
        assert isinstance(result, bool)
        assert result is True

    def test_get_horizontal_swing(self, controller):
        result = controller.get_horizontal_swing(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "manual3"

    async def test_set_horizontal_vane(self, controller):
        await controller.set_horizontal_vane(MOCK_DEVICE_ID, "manual4")

    def test_has_setpoint_control(self, controller):
        result = controller.has_setpoint_control(MOCK_DEVICE_ID)
        assert isinstance(result, bool)
        assert result is True

    def test_get_setpoint(self, controller):
        result = controller.get_setpoint(MOCK_DEVICE_ID)
        assert isinstance(result, float)
        assert result == 21.0

    def test_get_temperature(self, controller):
        result = controller.get_temperature(MOCK_DEVICE_ID)
        assert isinstance(result, float)
        assert result == 24.0

    async def test_set_temperature(self, controller):
        await controller.set_temperature(MOCK_DEVICE_ID, 10)

    def test_get_run_hours(self, controller):
        result = controller.get_run_hours(MOCK_DEVICE_ID)
        assert isinstance(result, int)
        assert result == MOCK_VAL_RUN_HOURS

    def test_get_error(self, controller):
        result = controller.get_error(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "H00: No abnormality detected"

    def test_get_min_setpoint(self, controller):
        result = controller.get_min_setpoint(MOCK_DEVICE_ID)
        assert isinstance(result, float)
        assert result == 18.0

    def test_get_max_setpoint(self, controller):
        result = controller.get_max_setpoint(MOCK_DEVICE_ID)
        assert isinstance(result, float)
        assert result == 30.0

    def test_get_outdoor_temperature(self, controller):
        result = controller.get_outdoor_temperature(MOCK_DEVICE_ID)
        assert isinstance(result, float)
        assert result == 26.0

    def test_get_preset_mode(self, controller):
        result = controller.get_preset_mode(MOCK_DEVICE_ID)
        assert isinstance(result, str)
        assert result == "eco"

    def test_get_devices(self, controller):
        result = controller.get_devices()
        assert isinstance(result, dict)
        assert len(result) == 1

    def test_get_device(self, controller):
        result = controller.get_device(MOCK_DEVICE_ID)
        assert isinstance(result, dict)
        assert len(result) > 20



#!/usr/bin/env python3
# read the contents of your README file
from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="pyintesishome2",
    version="1.8.9",
    description="A python3 library for running asynchronus communications with IntesisHome Smart AC Controllers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dolkensp/pyIntesisHome",
    author="James Nimmo",
    author_email="james@nimmo.net.nz",
    license="MIT",
    install_requires=["aiohttp>=3.7.4,<4"],
    packages=["pyintesishome2"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    ],
)



"""Main submodule for pyintesishome2."""

import asyncio
import logging
from json import JSONDecodeError

import aiohttp

from .const import (
    COMMAND_MAP,
    DEVICE_INTESISHOME_LOCAL,
    INTESIS_MAP,
    LOCAL_CMD_GET_AVAIL_DP,
    LOCAL_CMD_GET_DP_VALUE,
    LOCAL_CMD_GET_INFO,
    LOCAL_CMD_LOGIN,
    LOCAL_CMD_SET_DP_VALUE,
)
from .exceptions import IHAuthenticationError, IHConnectionError
from .intesisbase import IntesisBase

_LOGGER = logging.getLogger("pyintesishome2")


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-public-methods
class IntesisHomeLocal(IntesisBase):
    """pyintesishome2 local class."""

    def __init__(self, host, username, password, loop=None, websession=None) -> None:
        """Constructor"""
        device_type = DEVICE_INTESISHOME_LOCAL
        self._session_id: str = ""
        self._datapoints: dict = {}
        self._scan_interval = 6
        self._device_id: str = ""
        self._values: dict = {}
        self._info: dict = {}
        self._update_task = None
        super().__init__(
            host=host,
            username=username,
            password=password,
            loop=loop,
            websession=websession,
            device_type=device_type,
        )

    async def _request_values(self) -> dict:
        """Get all entity values."""
        response = {}
        try:
            response = await self._request(LOCAL_CMD_GET_DP_VALUE, uid="all")
        except (IHConnectionError, IHAuthenticationError) as exc:
            _LOGGER.error(
                "IntesisHome connection error: %s",
                exc,
            )

        if response and "dpval" in response:
            self._values = {dpval["uid"]: dpval["value"] for dpval in response["dpval"]}
            return self._values
        return {}

    async def _run_updater(self):
        """Run a loop that updates the values every _scan_interval."""
        try:
            while self._connected:
                try:
                    values = await self._request_values()
                    for uid, value in values.items():
                        self._update_device_state(self._device_id, uid, value)

                    await self._send_update_callback(self._device_id)
                except IHConnectionError as exc:
                    _LOGGER.error("Error during updater task: %s", exc)
                await asyncio.sleep(self._scan_interval)
        except asyncio.CancelledError:
            _LOGGER.debug("Cancelled the updater task")
        _LOGGER.debug("Updater task is exiting")

    async def _authenticate(self) -> bool:
        """Authenticate using username and password."""
        payload = {
            "command": LOCAL_CMD_LOGIN,
            "data": {"username": self._username, "password": self._password},
        }
        try:
            async with self._web_session.post(
                f"http://{self._host}/api.cgi", json=payload
            ) as response:
                if response.status != 200:
                    raise IHConnectionError(
                        "HTTP response status is unexpected (not 200)"
                    )
                json_response = await response.json()
                # Check if the response has the expected format
                if (
                    "data" in json_response
                    and "id" in json_response["data"]
                    and "sessionID" in json_response["data"]["id"]
                ):
                    self._session_id = json_response["data"]["id"]["sessionID"]
                    _LOGGER.debug(
                        "Authenticated with new session ID %s", self._session_id
                    )
                else:
                    _LOGGER.error("Unexpected response format during authentication")
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientResponseError,
            aiohttp.ClientPayloadError,
            aiohttp.ContentTypeError,
            JSONDecodeError,
        ) as exception:
            _LOGGER.error("Error during authentication: %s", str(exception))

    async def _request(self, command: str, **kwargs) -> dict:
        """Make a request."""
        connection_attempts = 0

        while connection_attempts < 2:
            connection_attempts += 1
            if not self._session_id:
                await self._authenticate()

            payload = {
                "command": command,
                "data": {"sessionID": self._session_id, **kwargs},
            }
            _LOGGER.debug(
                "Sending intesishome_local command %s to %s", command, self._host
            )
            timeout = aiohttp.ClientTimeout(total=10)
            json_response = {}
            try:
                async with self._web_session.post(
                    f"http://{self._host}/api.cgi",
                    json=payload,
                    timeout=timeout,
                ) as response:
                    if response.status != 200:
                        raise IHConnectionError(
                            f"HTTP response status is unexpected for {self._host}"
                            "(got {response.status}, want 200)"
                        )
                    json_response = await response.json()
            except asyncio.exceptions.TimeoutError as exc:
                _LOGGER.error(
                    "IntesisHome HTTP timeout error for %s: %s",
                    self._host,
                    exc,
                )
            except aiohttp.ClientError as exc:
                _LOGGER.error(
                    "IntesisHome HTTP error for %s: %s",
                    self._host,
                    exc,
                )

            # If the response has a non-false "success", return the data.
            #
            # If the key doesn't exist, treat it as "false" and continue with
            # error handling.
            #
            # If the error code is one we know requires re-authentication,
            # clear the session key so we re-authenticate on the next attempt,
            # otherwise just log the code and error message.
            #
            # If there's neither a "success" or "error" key, something is very
            # wonky, so log an error plus the entire response.
            if json_response.get("success", False):
                return json_response.get("data")
            if "error" in json_response:
                error = json_response["error"]
                if error.get("code") in [1, 5]:
                    self._session_id = ""
                    _LOGGER.debug(
                        "Request failed for %s (code=%s, message=%r)."
                        "Clearing session key to force re-authentication",
                        self._host,
                        error.get("code"),
                        error.get("message"),
                    )
                else:
                    _LOGGER.debug(
                        "Request failed for %s (code=%s, message=%r). Error not handled",
                        self._host,
                        error.get("code"),
                        error.get("message"),
                    )
            else:
                _LOGGER.debug(
                    "Request failed for %s - no 'success' or 'error' keys. json_response=%r",
                    self._host,
                    json_response,
                )

    async def _request_value(self, name: str) -> dict:
        """Get entity value by uid."""
        response = await self._request(
            LOCAL_CMD_GET_DP_VALUE, uid=COMMAND_MAP[name]["uid"]
        )
        return response["dpval"]["value"]

    async def _set_value(self, device_id, uid, value):
        return await self._request(
            LOCAL_CMD_SET_DP_VALUE,
            uid=uid,
            value=value,
        )

    async def get_datapoints(self) -> dict:
        """Get all available datapoints."""
        response = await self._request(LOCAL_CMD_GET_AVAIL_DP)
        self._datapoints = {
            dpoint["uid"]: dpoint for dpoint in response["dp"]["datapoints"]
        }
        if self._device_id:
            for dpoint in response["dp"]["datapoints"]:
                dpoint_name = INTESIS_MAP[dpoint["uid"]]["name"]
                self._devices[self._device_id][dpoint_name] = None
        return self._datapoints

    def _has_datapoint(self, datapoint: str):
        """Entity has a datapoint."""
        if datapoint not in COMMAND_MAP:
            return False
        return COMMAND_MAP[datapoint]["uid"] in self._datapoints

    async def connect(self):
        """Connect to the device and start periodic updater."""
        await self.poll_status()
        _LOGGER.debug("Successful authenticated and polled. Fetching Datapoints")
        await self.get_datapoints()
        self._connected = True
        _LOGGER.debug("Starting updater task")
        self._update_task = asyncio.create_task(self._run_updater())

    async def stop(self):
        """Disconnect and stop periodic updater."""
        _LOGGER.debug("Stopping updater task")
        await self._cancel_task_if_exists(self._update_task)
        self._connected = False

        if self._own_session:
            await self._web_session.close()

    async def get_info(self) -> dict:
        """Get device info."""
        response = await self._request(LOCAL_CMD_GET_INFO)
        self._info = response["info"]
        return self._info

    async def poll_status(self, sendcallback=False):
        """Get device info for setup purposes."""
        try:
            await self._authenticate()
            info = await self.get_info()

            # Extract device_id up to the first space, if there is a space
            raw_id = info.get("sn")
            if raw_id:
                device_id, *_ = raw_id.split(" ")
                self._device_id = device_id
                self._controller_id = device_id.lower()

            self._controller_name = (
                f"{self._info.get('deviceModel')} ({info.get('ownSSID')})"
            )
            # Setup devices
            self._devices[self._device_id] = {
                "name": info.get("ownSSID"),
                "widgets": [],
                "model": info.get("deviceModel"),
            }

            await self.get_datapoints()
            _LOGGER.debug(repr(self._devices))

            self._update_device_state(self._device_id, "acStatus", info.get("acStatus"))

            if sendcallback:
                await self._send_update_callback(str(self._device_id))
        except (IHConnectionError, KeyError) as exception:
            _LOGGER.error("Error during polling status: %s", str(exception))

    def get_mode_list(self, device_id) -> list:
        """Get possible entity modes."""
        uid = COMMAND_MAP["mode"]["uid"]
        return [
            INTESIS_MAP[uid]["values"][i]
            for i in self._datapoints[uid]["descr"]["states"]
        ]

    def _get_fan_map(self, device_id):
        fan_values = sorted(self._datapoints[4]["descr"]["states"])
        for values in INTESIS_MAP[67]["values"].values():
            if sorted(values.keys()) == fan_values:
                return values
        return INTESIS_MAP[67]["values"][63]

    def has_vertical_swing(self, device_id) -> bool:
        """Entity supports vertical swing."""
        return self._has_datapoint("vvane")

    def has_horizontal_swing(self, device_id) -> bool:
        """Entity supports horizontal swing."""
        return self._has_datapoint("hvane")

    async def _parse_response(self, decoded_data):
        """Private method to parse the API response."""
        raise NotImplementedError()
